@@ -5,6 +5,7 @@
 #include <QImage>
 
 class QTemporaryDir;
+class QFile;
 
 #define ARRAY_SIZE(array) \
   (sizeof(array) / sizeof(array[0]))
@@ -24,6 +25,7 @@ namespace Util
 	QString displayName(const QString &path);
 	const QTemporaryDir &tempDir();
 	bool isSQLiteDB(const QString &path, QString &errorString);
+	bool isZIP(QFile *file);
 	QImage svg2img(const QString &path, qreal ratio);
 	QByteArray gunzip(const QByteArray &data);
 }
