@@ -14,21 +14,12 @@
 #include <QFileInfo>
 #include <QTemporaryDir>
 #include <QCryptographicHash>
-#include <QtEndian>
 #include <QUrl>
 #include <QRegularExpression>
 #include <private/qzipreader_p.h>
 #include "common/util.h"
 #include "kmlparser.h"
 
-static bool isZIP(QFile *file)
-{
-	quint32 magic;
-
-	return (file->read((char *)&magic, sizeof(magic)) == (qint64)sizeof(magic)
-	  && qFromLittleEndian(magic) == 0x04034b50);
-}
-
 qreal KMLParser::number()
 {
 	bool res;
@@ -541,7 +532,7 @@ void KMLParser::photoOverlay(const Ctx &ctx, QVector<Waypoint> &waypoints,
 	Waypoint w;
 	QMap<QString, PolygonStyle> unused;
 	QMap<QString, LineStyle> unused2;
-	static QRegularExpression re("\\$\\[[^\\]]+\\]");
+	static const QRegularExpression re("\\$\\[[^\\]]+\\]");
 
 	while (_reader.readNextStartElement()) {
 		if (_reader.name() == QLatin1String("name"))
@@ -934,8 +925,8 @@ bool KMLParser::parse(QFile *file, QList<TrackData> &tracks,
 
 	_reader.clear();
 
-	if (isZIP(file)) {
-		QZipReader zip(fi.absoluteFilePath(), QIODevice::ReadOnly);
+	if (Util::isZIP(file)) {
+		QZipReader zip(file);
 		QTemporaryDir tempDir;
 		if (!tempDir.isValid() || !zip.extractAll(tempDir.path()))
 			_reader.raiseError("Error extracting ZIP file");
@@ -964,7 +955,6 @@ bool KMLParser::parse(QFile *file, QList<TrackData> &tracks,
 			}
 		}
 	} else {
-		file->reset();
 		_reader.setDevice(file);
 
 		if (_reader.readNextStartElement()) {
