@@ -8,17 +8,12 @@
 class AngularUnits
 {
 public:
-	AngularUnits() : _code(0), _f(NAN) {}
+	AngularUnits() : _f(NAN) {}
 	AngularUnits(int code);
-	AngularUnits(double val) : _code(0), _f(rad2deg(val)) {}
+	AngularUnits(double val) : _f(rad2deg(val)) {}
 
 	bool operator==(const AngularUnits &other) const
-	{
-		if (_code == 9110)
-			return (other._code == 9110);
-		else
-			return (_f == other._f);
-	}
+	  {return fabs(_f - other._f) < 1e-6;}
 
 	bool isNull() const {return std::isnan(_f);}
 	bool isValid() const {return !std::isnan(_f);}
@@ -29,8 +24,8 @@ public:
 #ifndef QT_NO_DEBUG
 	friend QDebug operator<<(QDebug dbg, const AngularUnits &au);
 #endif // QT_NO_DEBUG
+
 private:
-	int _code;
 	double _f;
 };
 
