@@ -11,13 +11,15 @@ public:
 	PrimeMeridian(int code);
 	PrimeMeridian(double lon) : _pm(lon) {}
 
+	bool operator==(const PrimeMeridian &other) const
+	  {return fabs(_pm - other._pm) < 1e-6;}
+
 	bool isNull() const {return std::isnan(_pm);}
 	bool isValid() const {return !std::isnan(_pm);}
 
 	double toGreenwich(double val) const;
 	double fromGreenwich(double val) const;
 
-	friend bool operator==(const PrimeMeridian &pm1, const PrimeMeridian &pm2);
 #ifndef QT_NO_DEBUG
 	friend QDebug operator<<(QDebug dbg, const PrimeMeridian &pm);
 #endif // QT_NO_DEBUG
@@ -26,9 +28,6 @@ private:
 	double _pm;
 };
 
-inline bool operator==(const PrimeMeridian &pm1, const PrimeMeridian &pm2)
-  {return (pm1._pm == pm2._pm);}
-
 #ifndef QT_NO_DEBUG
 QDebug operator<<(QDebug dbg, const PrimeMeridian &pm);
 #endif // QT_NO_DEBUG
